/* The following copyright applies to: 					*/
/* CaptureReset()							*/

/*
 * Copyright (c) 2002
 * Ningning Hu and the Carnegie Mellon University.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* The following copyright applies to: 					*/
/* pcap_lookupdev_withIP()						*/

/*
 * Copyright (c) 1994, 1995, 1996, 1997, 1998
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Systems
 *      Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* The following copyright applies to all the code expcept:		*/ 
/* CaptureReset() and pcap_lookupdev_withIP()				*/

/*
 * Copyright (c) 1999
 * Stefan Savage and the University of Washington.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <sys/types.h>
#include <sys/param.h>
#include <string.h>
#include <stdio.h>
#include "gmt2local.h"
#include "pcap.h"
#include "inet.h"
#include "capture.h"

#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <errno.h>
#include <memory.h>
#include <arpa/inet.h>

#ifdef IFF_LOOPBACK
#define ISLOOPBACK(p) ((p)->ifr_flags & IFF_LOOPBACK)
#else
#define ISLOOPBACK(p) ((p)->ifr_name[0] == 'l' && (p)->ifr_name[1] == 'o' && \
		    (isdigit((p)->ifr_name[2]) || (p)->ifr_name[2] == '\0'))
#endif


#define DEFAULT_SNAPLEN 80

pcap_t *pc;		/* pcap device */
int datalinkOffset;	/* offset of ip packet from datalink packet */

int captureDebug = 0;

unsigned int thisTimeZone;

/* This function is modified from libpcap-0.4/inet.c */
char * pcap_lookupdev_withIP(uint32 dstIP, register char * errbuf)
{
        register int fd, minunit, n;
        register char *cp;
        register struct ifreq *ifrp, *ifend, *ifnext, *mp;
        struct ifconf ifc;
        struct ifreq ibuf[64], ifr;
        static char device[sizeof(ifrp->ifr_name) + 1];

	struct sockaddr_in *sin;

        fd = socket(AF_INET, SOCK_DGRAM, 0);
        if (fd < 0) {
                (void)sprintf(errbuf, "socket: %s", pcap_strerror(errno));
                return (NULL);
        }
        ifc.ifc_len = sizeof ibuf;
        ifc.ifc_buf = (caddr_t)ibuf;

        memset((char *)ibuf, 0, sizeof(ibuf));
        if (ioctl(fd, SIOCGIFCONF, (char *)&ifc) < 0 ||
            ifc.ifc_len < sizeof(struct ifreq)) {
                (void)sprintf(errbuf, "SIOCGIFCONF: %s", pcap_strerror(errno));
                //(void)close(fd);
                close(fd);
                return (NULL);
        }
        ifrp = ibuf;
        ifend = (struct ifreq *)((char *)ibuf + ifc.ifc_len);

        mp = NULL;
        minunit = 666;
        for (; ifrp < ifend; ifrp = ifnext) {
#ifdef HAVE_SOCKADDR_SA_LEN
                n = ifrp->ifr_addr.sa_len + sizeof(ifrp->ifr_name);
                if (n < sizeof(*ifrp))
                        ifnext = ifrp + 1;
                else
                        ifnext = (struct ifreq *)((char *)ifrp + n);
                if (ifrp->ifr_addr.sa_family != AF_INET)
                        continue;
#else
                ifnext = ifrp + 1;
#endif
                /*
                 * Need a template to preserve address info that is
                 * used below to locate the next entry.  (Otherwise,
                 * SIOCGIFFLAGS stomps over it because the requests
                 * are returned in a union.)
                 */
                strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
                if (ioctl(fd, SIOCGIFFLAGS, (char *)&ifr) < 0) {
                        if (errno == ENXIO)
                                continue;
                        (void)sprintf(errbuf, "SIOCGIFFLAGS: %.*s: %s",
                            (int)sizeof(ifr.ifr_name), ifr.ifr_name,
                            pcap_strerror(errno));
                        close(fd);
                        return (NULL);
                }

                /* Must be up and not the loopback */
                if ((ifr.ifr_flags & IFF_UP) == 0 || ISLOOPBACK(&ifr))
                        continue;

                for (cp = ifrp->ifr_name; !isdigit(*cp); ++cp)
                        continue;
                n = atoi(cp);
#if 0
                if (n < minunit) {
#else
                strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
        	ifr.ifr_addr.sa_family = AF_INET;
        	if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
		    sin = (struct sockaddr_in *)&(ifr.ifr_addr);
		}
		/*dst_in.s_addr = dstIP;
		printf("%s %s %s\n", 
			ifrp->ifr_name, 
			inet_ntoa(sin->sin_addr),
			inet_ntoa(dst_in));
		*/
                if ((n < minunit) && 
		    (sin->sin_addr.s_addr == dstIP)) {
#endif
                        minunit = n;
                        mp = ifrp;
                }
        }
        (void)close(fd);
        if (mp == NULL) {
                (void)strcpy(errbuf, "no suitable device found");
                return (NULL);
        }

        (void)strncpy(device, mp->ifr_name, sizeof(device) - 1);
        device[sizeof(device) - 1] = '\0';
        return (device);
}


pcap_t * CaptureInit_withDev(char* device, uint32 sourceIP, uint16 sourcePort,
		 uint32 targetIP, uint16 targetPort)
{
  // char *device;
  char errbuf[PCAP_ERRBUF_SIZE];
  int snaplen = DEFAULT_SNAPLEN;
  int promisc = 1;
  int timeout = 10;  /* timeout in 1 second (10 ms) */
  char filtercmds[255];
  bpf_u_int32 netmask;
  struct bpf_program filter;
  char source[18];
  char target[18];
  int i;

  /* Get local time zone for interpreting timestamps */
  /* XXX - does this belong here? */
  thisTimeZone = gmt2local(0);

  // device = pcap_lookupdev(errbuf);
  /* device = pcap_lookupdev_withIP(targetIP, errbuf);
  if (device == NULL) {
    fprintf(stderr, "Can't find capture device: %s\n", errbuf);
    exit(-1);
  }*/
  if (captureDebug) {
    printf("Device name is %s\n", device);
  }
  pc = pcap_open_live(device, snaplen, promisc, timeout, errbuf);
  if (pc == NULL) {
    fprintf(stderr,"Can't open capture device %s: %s\n",device, errbuf);
    exit(-1);
  } 

  /* XXX why do we need to do this? */
  i = pcap_snapshot(pc);
  if (snaplen < i) {
    fprintf(stderr, "Warning: snaplen raised to %d from %d",
	    snaplen, i);
  }

  if ((i = pcap_datalink(pc)) < 0) {
    fprintf(stderr,"Unable to determine datalink type for %s: %s\n",
	    device, errbuf);
    exit(-1);
  }
  switch(i) {
    case DLT_EN10MB: datalinkOffset = 14; break;
    case DLT_IEEE802: datalinkOffset = 22; break;
    case DLT_NULL: datalinkOffset = 4; break;
    case DLT_SLIP: 
    case DLT_PPP: datalinkOffset = 24; break;
    case DLT_RAW: datalinkOffset = 0; break;
    default: 
       fprintf(stderr,"Unknown datalink type %d\n",i);
       exit(-1);
       break;
  }

  if (InetAddress(sourceIP) < 0) {
    fprintf(stderr, "Invalid source IP address (%d)\n", sourceIP);
    exit(-1);
  }

  strncpy(source, InetAddress(sourceIP), sizeof(source)-1);
  strncpy(target, InetAddress(targetIP), sizeof(target)-1);


  /* Setup initial filter */
  /* sprintf(filtercmds,
	  "host %s and host %s and port %d and port %d\n",
	  source, target, sourcePort, targetPort);
  */

  sprintf(filtercmds,
	  // "host %s and host %s and port %d\n",
	  "src host %s and dst host %s\n",
	  source, target);
	  // source, target, targetPort);

  if (captureDebug) {
    printf("datalinkOffset = %d\n", datalinkOffset);
    printf("filter = %s\n", filtercmds);
  }
  if (pcap_compile(pc, &filter, filtercmds, 1, netmask) < 0) {
    printf("Error: %s", pcap_geterr(pc));
    exit(-1);
  }

  if (pcap_setfilter(pc, &filter) < 0) {
    fprintf(stderr, "Can't set filter: %s",pcap_geterr(pc));
    exit(-1);
  }
  
  if (captureDebug) {
    printf("Listening on %s...\n", device);
  }

  return pc;
}


void CaptureReset(pcap_t * pc,
                 uint32 * sourceIP, uint32 * sourcePort,
                 uint32 * targetIP, uint32 * targetPort,
		 int count)
{
	char filtercmds[1024];
	bpf_u_int32 netmask;
	struct bpf_program filter;
	char source[18];
	char target[18];
	int len;
	int i;

	len = 0;
  	for (i=0; i<count; i++) {
   	    strncpy(source, InetAddress(sourceIP[i]), sizeof(source)-1);
    	    strncpy(target, InetAddress(targetIP[i]), sizeof(target)-1);

	    if (i>0) {
  	       sprintf(filtercmds + len, " or ");
	       len = strlen(filtercmds);
	    }

  	    sprintf(filtercmds + len, " ( src host %s and dst host %s\n ) ",
          	source, target);
	    len = strlen(filtercmds);
  	}

	if (captureDebug) {
	    printf("filtercmds: %s\n", filtercmds);
	} 

	if (pcap_compile(pc, &filter, filtercmds, 1, netmask) < 0) {
	    printf("Error: %s", pcap_geterr(pc));
	    exit(-1);
	} 

	if (pcap_setfilter(pc, &filter) < 0) {
	    fprintf(stderr, "Can't set filter: %s",pcap_geterr(pc));
	    exit(-1);
	} 
}


char *CaptureGetPacket(pcap_t *pc, struct PacketInfo *pi)
{
  const u_char *p;
  
  p = pcap_next(pc, (struct pcap_pkthdr *)pi);
  if (p!= NULL) {
    p += datalinkOffset;
  }
  pi->ts.tv_sec = (pi->ts.tv_sec + thisTimeZone) % 86400;
  return (char *)p;
}


void CaptureEnd(pcap_t * pc)
{
  struct pcap_stat stat;

  if (pcap_stats(pc, &stat) < 0) {
    // (void)fprintf(stderr, "pcap_stats: %s\n", pcap_geterr(pc));
  }
  else {
    // (void)fprintf(stderr, "%d packets received by filter\n", stat.ps_recv); 
    // (void)fprintf(stderr, "%d packets dropped by kernel\n", stat.ps_drop);
  }

  pcap_close(pc);
}

